import os, sys
import numpy

from PyQt5.QtCore import QRect, Qt
from PyQt5.QtWidgets import QApplication, QMessageBox, QLabel, QSizePolicy
from PyQt5.QtGui import QTextCursor, QPixmap

import orangecanvas.resources as resources

from orangewidget import gui, widget
from orangewidget.settings import Setting

from oasys.widgets.widget import OWWidget
from oasys.widgets import gui as oasysgui


from orangecontrib.esrf.shadow.util.wolter1 import recipe1, recipe2, recipe3, recipe4, rotate_and_shift_quartic
from orangecontrib.esrf.shadow.util.shadow_objects import ConicCoefficientsPreProcessorData

from oasys.util.oasys_util import EmittingStream

class OWWolterCalculator(OWWidget):
    name = "Wolter Calculator"
    id = "WolterCalculator"
    description = "Calculation of coefficients for Wolter systems"
    icon = "icons/wolter2.png"
    author = "Manuel Sanchez del Rio"
    maintainer_email = "srio@esrf.eu"
    priority = 7
    category = ""
    keywords = ["oasys", "wolter", "telescope", "advanced KB"]

    outputs = [{"name":"ConicCoeff_1_PreProcessor_Data",
                "type":ConicCoefficientsPreProcessorData,
                "doc":"ConicCoeff #1 PreProcessor Data",
                "id":"ConicCoeff_1_PreProcessor_Data"},
               {"name": "ConicCoeff_2_PreProcessor_Data",
                "type": ConicCoefficientsPreProcessorData,
                "doc": "ConicCoeff #2 PreProcessor Data",
                "id": "ConicCoeff_2_PreProcessor_Data"},
               ]

    want_main_area = True
    want_control_area = True

    MAX_WIDTH = 1320
    MAX_HEIGHT = 700

    IMAGE_WIDTH = 860
    IMAGE_HEIGHT = 645

    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT = 650 #18

    #################

    setup_type = Setting(0)
    same_angle = Setting(1)

    p1 = Setting(2.1)
    q1 = Setting(1.05)
    p2 = Setting(1.904995)
    q2 = Setting(10.0)
    distance = Setting(0.1)

    theta1 = Setting(0.1134464014)
    theta2 = Setting(0.003)

    ratio_hyp = Setting(1.1)  # ratio_hyp = q_hyp / p_ell > 1.0
    m_hyp = Setting(1 / 3)

    # to send to shadow
    conic_coefficients1 = Setting([0] * 10)
    conic_coefficients2 = Setting([0] * 10)
    source_plane_distance1 = Setting(0.0)
    source_plane_distance2 = Setting(0.0)
    image_plane_distance1 = Setting(0.0)
    image_plane_distance2 = Setting(0.0)
    angles_respect_to1 = Setting(0)
    angles_respect_to2 = Setting(0)
    incidence_angle_deg1 = Setting(0.0)
    incidence_angle_deg2 = Setting(0.0)
    reflection_angle_deg1 = Setting(0.0)
    reflection_angle_deg2 = Setting(0.0)
    mirror_orientation_angle1 = Setting(0)
    mirror_orientation_angle2 = Setting(0)

    npoints = Setting(200)
    y_length = Setting(0.3)

    sagittal_flat = Setting(0)

    tab=[]

    usage_path = os.path.join(resources.package_dirname("orangecontrib.esrf.shadow.widgets.extension"), "icons", "use_wolter.png")

    def __init__(self):
        super().__init__()

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width() * 0.05),
                               round(geom.height() * 0.05),
                               round(min(geom.width() * 0.98, self.MAX_WIDTH)),
                               round(min(geom.height() * 0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        gui.separator(self.controlArea)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_calc = oasysgui.createTabPage(tabs_setting, "Calculate")
        # tab_out = oasysgui.createTabPage(tabs_setting, "Plot")
        tab_usa = oasysgui.createTabPage(tabs_setting, "Use of the Widget")


        #
        #-------------------- calculate
        #

        button = gui.button(tab_calc, self, "Calculate", callback=self.calculate)

        tab_step_1 = oasysgui.widgetBox(tab_calc, "Calculation Parameters", addSpace=True, orientation="vertical", height=600)

        box = oasysgui.widgetBox(tab_step_1, "setup inputs", orientation="vertical")

        gui.comboBox(box, self, "setup_type", label="Setup type", labelWidth=260,
                     items=["Wolter-I variable throw",
                            "Wolter-I fixed throw",
                            ],
                     callback=self.update_panel, sendSelectedValue=False, orientation="horizontal")

        self.w_p1 = oasysgui.lineEdit(box, self, "p1", "Distance focus11-oe1 (p1) [m]", labelWidth=210, valueType=float, orientation="horizontal")
        self.w_q1 = oasysgui.lineEdit(box, self, "q1", "Distance oe1-focus12 (q1) [m]", labelWidth=210, valueType=float, orientation="horizontal")
        self.w_p2 = oasysgui.lineEdit(box, self, "p2", "Distance focus21-oe2 (p2) [m]", labelWidth=210, valueType=float, orientation="horizontal")
        self.w_q2 = oasysgui.lineEdit(box, self, "q2", "Distance oe2-focus22 (q2) [m]", labelWidth=210, valueType=float, orientation="horizontal")
        self.w_distance = oasysgui.lineEdit(box, self, "distance", "Distance oe1-oe2 [m]", labelWidth=210, valueType=float, orientation="horizontal")

        self.w_ratio_hyp = oasysgui.lineEdit(box, self, "ratio_hyp", "Ratio hyperbola=q2/p2>1", labelWidth=210, valueType=float, orientation="horizontal")
        self.w_m_hyp = oasysgui.lineEdit(box, self, "m_hyp", "Magnification hyperbola=p2/q2", labelWidth=220,
                                             valueType=float, orientation="horizontal")

        gui.comboBox(box, self, "sagittal_flat", label="2D/1D focusing", labelWidth=260,
                     items=["2D (revolution symmetry)",
                            "1D (flat in sagittal)",
                            ], sendSelectedValue=False, orientation="horizontal")

        gui.separator(box)

        box_2 = oasysgui.widgetBox(tab_step_1, "angles", orientation="vertical")

        gui.comboBox(box_2, self, "same_angle", label="Same grazing angles", labelWidth=260,
                     items=["No",
                            "Yes",
                            ],
                     callback=self.update_panel, sendSelectedValue=False, orientation="horizontal")

        self.w_theta1 = oasysgui.lineEdit(box_2, self, "theta1", "Grazing angle oe1 [rad]", labelWidth=210, valueType=float, orientation="horizontal", callback=self.update_panel)
        self.w_theta2 = oasysgui.lineEdit(box_2, self, "theta2", "Grazing angle oe2 [rad]", labelWidth=210, valueType=float, orientation="horizontal")



        box_3 = oasysgui.widgetBox(tab_step_1, "For plotting mirror profiles", orientation="vertical")
        oasysgui.lineEdit(box_3, self, "npoints", "Points", labelWidth=260, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(box_3, self, "y_length", "Mirror length [m]", labelWidth=260, valueType=int, orientation="horizontal")

        #
        #-------------------- Use
        #

        tab_usa.setStyleSheet("background-color: white;")

        usage_box = oasysgui.widgetBox(tab_usa, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.usage_path))

        usage_box.layout().addWidget(label)

        #
        #
        #

        gui.rubber(self.controlArea)

        self.initializeTabs()

        self.update_panel()

        gui.rubber(self.mainArea)

    def update_panel(self):
        self.w_p1.setVisible(True)
        self.w_q1.setVisible(True)
        self.w_p2.setVisible(True)
        self.w_q2.setVisible(True)
        self.w_theta1.setVisible  (True)
        self.w_theta2.setVisible  (True)
        self.w_distance.setVisible(True)

        self.w_ratio_hyp.setEnabled(False)
        self.w_m_hyp.setEnabled(False)

        if self.setup_type == 0:
            self.w_p1.setEnabled(True)
            self.w_q1.setEnabled(True)
            self.w_p2.setEnabled(False)
            self.w_q2.setEnabled(False)
            self.w_theta1.setEnabled(True)
            self.w_theta2.setEnabled(True)
            self.w_distance.setEnabled(True)
        elif self.setup_type == 1:
            self.w_p1.setEnabled(True)
            self.w_q1.setEnabled(False)
            self.w_p2.setEnabled(True)
            self.w_q2.setEnabled(False)
            self.w_theta1.setEnabled(True)
            self.w_theta2.setEnabled(True)
            self.w_distance.setEnabled(True)
        elif self.setup_type == 2:
            self.w_p1.setEnabled(True)
            self.w_p2.setEnabled(True)
            self.w_p2.setEnabled(True)
            self.w_q2.setEnabled(False)
            self.w_theta1.setEnabled(True)
            self.same_angle = True
            self.w_distance.setEnabled(False)
        elif self.setup_type == 3:
            self.w_p1.setEnabled(True)
            self.w_q1.setEnabled(False)
            self.w_p2.setEnabled(True)
            self.w_q2.setEnabled(False)
            self.w_theta1.setEnabled(True)
            self.same_angle = True
            self.w_distance.setEnabled(False)
        else:
            raise Exception(NotImplementedError)


        if self.same_angle:
            self.theta2 = self.theta1
            self.w_theta1.setEnabled(True)
            self.w_theta2.setEnabled(False)
        else:
            self.w_theta1.setEnabled(True)
            self.w_theta2.setEnabled(True)

        if self.setup_type == 0:
            self.w_ratio_hyp.setEnabled(True)
            self.w_m_hyp.setEnabled(False)
        elif self.setup_type == 1:
            self.w_ratio_hyp.setEnabled(False)
            self.w_m_hyp.setEnabled(True)
        elif self.setup_type == 2:
            self.w_ratio_hyp.setEnabled(False)
            self.w_m_hyp.setEnabled(False)
            self.q2 = self.q1
            self.distance = 0.0
        elif self.setup_type == 3:
            self.w_ratio_hyp.setEnabled(False)
            self.w_m_hyp.setEnabled(False)
            self.q1 = 0
            self.q2 = 0
            self.distance = 0.0
        else:
            raise Exception(NotImplementedError)


    def calculate(self):
        try:
            self.shadow_output.setText("")
            self.design_output.setText("")

            sys.stdout = EmittingStream(textWritten=self.writeStdOut)

            self.check_fields()

            results_txt = ""
            print("####################################################")
            print("# DESIGN PHASE")
            print("####################################################\n")

            if self.same_angle:
                theta_hyp = self.theta1
            else:
                theta_hyp = self.theta2

            if self.setup_type == 0:

                print("||||||||||||||||| theta_hyp: ", theta_hyp)
                tkt_ell, tkt_hyp = recipe1(
                    p_ell=self.p1,
                    q_ell=self.q1,
                    distance=self.distance,
                    theta=self.theta1,
                    theta_hyp=theta_hyp,
                    ratio_hyp=self.ratio_hyp,
                )
                print("||||||||||||||||| tkt_hyp: ", tkt_hyp)
                print("\n\n>>>>>\n\n")

                # correct for incidence in the negative Y
                print(">>> corrected hyperbola for incidence in the negative Y ")
                ccc1 = tkt_hyp['ccc']
                ccc2 = rotate_and_shift_quartic(ccc1, omega=0.0, theta=0.0, phi=numpy.pi, )
                tkt_hyp['ccc'] = ccc2
                print(ccc2)

                # 1D focusing (flat in sagittal)
                # xx  yy  zz  xy  yz  xz  x   y   z   0
                #  0   1   2   3   4   5  6   7   8   9
                if self.sagittal_flat:
                    print(">>> corrected: flat in sagittal (1D focusing)")
                    tkt_ell['ccc'][0] = 0
                    tkt_ell['ccc'][3] = 0
                    tkt_ell['ccc'][5] = 0
                    tkt_ell['ccc'][6] = 0

                    tkt_hyp['ccc'][0] = 0
                    tkt_hyp['ccc'][3] = 0
                    tkt_hyp['ccc'][5] = 0
                    tkt_hyp['ccc'][6] = 0

                # round
                for i in range(10):
                    if numpy.abs(tkt_ell['ccc'][i]) < 1e-14: tkt_ell['ccc'][i] = 0
                    if numpy.abs(tkt_hyp['ccc'][i]) < 1e-14: tkt_hyp['ccc'][i] = 0

                self.p2     = tkt_hyp['p']
                self.q2     = tkt_hyp['q']
                self.theta2 = tkt_hyp['theta_grazing']
                self.m_hyp  = 1/self.ratio_hyp

                print(tkt_ell)
                print(tkt_hyp)

                source_plane_distance_1 = self.p1
                image_plane_distance_1 = self.distance / 2
                angles_respect_to_1 = 0
                incidence_angle_deg_1 = 90 - numpy.degrees(self.theta1)
                reflection_angle_deg_1 = 90 - numpy.degrees(self.theta1)
                mirror_orientation_angle_1 = None

                source_plane_distance_2 = self.distance / 2
                image_plane_distance_2 = self.p2
                angles_respect_to_2 = 0
                incidence_angle_deg_2 = 90 - numpy.degrees(theta_hyp)
                reflection_angle_deg_2 =  90 - numpy.degrees(theta_hyp)
                mirror_orientation_angle_2 = None

                # round (cosmetics)
                self.p2     = numpy.round(self.p2    , 5)
                self.q2     = numpy.round(self.q2    , 5)
                # self.theta2 = numpy.round(self.theta2, 5)
                self.m_hyp  = numpy.round(self.m_hyp , 5)


            elif self.setup_type == 1:

                tkt_ell, tkt_hyp = recipe2(
                    p_ell=self.p1,
                    distance=self.distance,
                    p_hyp=self.p2,
                    theta=self.theta1,
                    theta_hyp=theta_hyp,
                    m_hyp=self.m_hyp,
                    verbose=1,
                )

                print("\n\n>>>>>\n\n")

                # correct for incidence in the negative Y
                print(">>> corrected hyperbola for incidence in the negative Y ")
                ccc1 = tkt_hyp['ccc']
                ccc2 = rotate_and_shift_quartic(ccc1, omega=0.0, theta=0.0, phi=numpy.pi, )
                tkt_hyp['ccc'] = ccc2
                print(ccc2)

                # 1D focusing (flat in sagittal)
                if self.sagittal_flat:
                    print(">>> corrected: flat in sagittal (1D focusing)")

                    tkt_ell['ccc'][0] = 0
                    tkt_ell['ccc'][3] = 0
                    tkt_ell['ccc'][5] = 0
                    tkt_ell['ccc'][6] = 0

                    tkt_hyp['ccc'][0] = 0
                    tkt_hyp['ccc'][3] = 0
                    tkt_hyp['ccc'][5] = 0
                    tkt_hyp['ccc'][6] = 0

                # round
                for i in range(10):
                    if numpy.abs(tkt_ell['ccc'][i]) < 1e-14: tkt_ell['ccc'][i] = 0
                    if numpy.abs(tkt_hyp['ccc'][i]) < 1e-14: tkt_hyp['ccc'][i] = 0

                self.q1 = tkt_ell['p']
                self.q2 = tkt_hyp['q']
                self.theta2 = tkt_hyp['theta_grazing']
                self.rat_hyp = 1/self.m_hyp

                print(tkt_ell)
                print(tkt_hyp)


                source_plane_distance_1 = self.p1
                image_plane_distance_1 = self.distance / 2
                angles_respect_to_1 = 0
                incidence_angle_deg_1 = 90 - numpy.degrees(self.theta1)
                reflection_angle_deg_1 = 90 - numpy.degrees(self.theta1)
                mirror_orientation_angle_1 = None

                if self.same_angle:
                    theta_hyp = self.theta1
                else:
                    theta_hyp = self.theta2

                source_plane_distance_2 = self.distance / 2
                image_plane_distance_2 = self.p2
                angles_respect_to_2 = 0
                incidence_angle_deg_2 = 90 - numpy.degrees(theta_hyp)
                reflection_angle_deg_2 =  90 - numpy.degrees(theta_hyp)
                mirror_orientation_angle_2 = None

                # round (cosmetics)
                self.p2     = numpy.round(self.p2    , 5)
                self.q2     = numpy.round(self.q2    , 5)
                # self.theta2 = numpy.round(self.theta2, 5)
                self.m_hyp  = numpy.round(self.m_hyp , 5)

            else:
                raise Exception(NotImplementedError)

            try:
                results_txt += "\nellipse a=%f" % tkt_ell['a']
                results_txt += "\nellipse b=%f" % tkt_ell['b']
                results_txt += "\nellipse c=%f" % tkt_ell['c']
                results_txt += "\nhyperbola a=%f" % tkt_hyp['a']
                results_txt += "\nhyperbola b=%f" % tkt_hyp['b']
                results_txt += "\nhyperbola c=%f" % tkt_hyp['c']
            except:
                pass

            ccc_ell = tkt_ell['ccc']
            ccc_hyp = tkt_hyp['ccc']

            self.conic_coefficients1 = ccc_ell
            self.conic_coefficients2 = ccc_hyp

            # ccc_hyp = rotate_and_shift_quartic(ccc_hyp, omega=0.0, theta=0.0, phi=numpy.pi, )
            if self.sagittal_flat:
                ccc_ell_nor = ccc_ell[1]
                ccc_hyp_nor = ccc_hyp[1]
            else:
                ccc_ell_nor = ccc_ell[0]
                ccc_hyp_nor = ccc_hyp[0]

            results_txt += "\n\n\n    oe1(normalized)      oe2(normalized)"
            for i in range(10):
                results_txt += "\nccc[%d]       %10.4g       %10.4g  " % (i,
                                                                    ccc_ell[i]/ccc_ell_nor, ccc_hyp[i]/ccc_hyp_nor)

            results_txt += "\n\n\n    oe1           oe2 "
            for i in range(10):
                results_txt += "\nccc[%d]       %10.4g       %10.4g  " % (i, ccc_ell[i], ccc_hyp[i])

            #results_txt += "\nthrow=%f m" % (self.p1+self.distance+self.p2)
            self.design_output.setText(results_txt)

            #
            # plot data
            #

            self.progressBarInit()

            # plot oe 1
            y, z1a, z1b = self.height(oe=1)
            y, z2a, z2b = self.height(oe=2)
            self.plot_multi_data1D([y,y,y,y], [z1a,z1b,z2a,z2b],
                                  10, 2, 0,
                                  title="mirror 1", xtitle="y [m]", ytitle="z [m]",
                                  ytitles=["Mirror 1 solution 1","Mirror 1 solution 2","Mirror 2 solution 1","Mirror 2 solution 2"],
                                  colors=['blue','red','green','k'],
                                  replace=True,
                                  control=False,
                                  xrange=None,
                                  yrange=None,
                                  symbol=['','','',''],
                                )

            self.progressBarFinished()
            #
            # send data
            #
            print("\n\n\n\n")
            print("####################################################")
            print("# RAY-TRACING PHASE")
            print("####################################################\n")

            print("---------------------- first mirror:")
            print("sending ccc ell: ", self.conic_coefficients1)
            if source_plane_distance_1    is not None: print("sending source_plane_distance=", source_plane_distance_1)
            if image_plane_distance_1     is not None: print("sending image_plane_distance=", image_plane_distance_1)
            if angles_respect_to_1        is not None: print("sending angles_respect_to=", angles_respect_to_1)
            if incidence_angle_deg_1      is not None: print("sending incidence_angle_deg=", incidence_angle_deg_1)
            if reflection_angle_deg_1     is not None: print("sending reflection_angle_deg=", reflection_angle_deg_1)
            if mirror_orientation_angle_1 is not None: print("sending mirror_orientation_angle=", mirror_orientation_angle_1)

            print("---------------------- second mirror:")
            print("sending ccc hyp: ", self.conic_coefficients2)
            if source_plane_distance_2    is not None: print("sending source_plane_distance=", source_plane_distance_2)
            if image_plane_distance_2     is not None: print("sending image_plane_distance=", image_plane_distance_2)
            if angles_respect_to_2        is not None: print("sending angles_respect_to=", angles_respect_to_2)
            if incidence_angle_deg_2      is not None: print("sending incidence_angle_deg=", incidence_angle_deg_2)
            if reflection_angle_deg_2     is not None: print("sending reflection_angle_deg=", reflection_angle_deg_2)
            if mirror_orientation_angle_2 is not None: print("sending mirror_orientation_angle=", mirror_orientation_angle_2)

            self.send("ConicCoeff_1_PreProcessor_Data", ConicCoefficientsPreProcessorData(
                conic_coefficient_0 = self.conic_coefficients1[0],
                conic_coefficient_1 = self.conic_coefficients1[1],
                conic_coefficient_2 = self.conic_coefficients1[2],
                conic_coefficient_3 = self.conic_coefficients1[3],
                conic_coefficient_4 = self.conic_coefficients1[4],
                conic_coefficient_5 = self.conic_coefficients1[5],
                conic_coefficient_6 = self.conic_coefficients1[6],
                conic_coefficient_7 = self.conic_coefficients1[7],
                conic_coefficient_8 = self.conic_coefficients1[8],
                conic_coefficient_9 = self.conic_coefficients1[9],
                source_plane_distance = source_plane_distance_1,
                image_plane_distance = image_plane_distance_1,
                angles_respect_to = angles_respect_to_1,
                incidence_angle_deg = incidence_angle_deg_1,
                reflection_angle_deg = reflection_angle_deg_1,
                mirror_orientation_angle = mirror_orientation_angle_1,
                ))
            self.send("ConicCoeff_2_PreProcessor_Data", ConicCoefficientsPreProcessorData(
                conic_coefficient_0=self.conic_coefficients2[0],
                conic_coefficient_1=self.conic_coefficients2[1],
                conic_coefficient_2=self.conic_coefficients2[2],
                conic_coefficient_3=self.conic_coefficients2[3],
                conic_coefficient_4=self.conic_coefficients2[4],
                conic_coefficient_5=self.conic_coefficients2[5],
                conic_coefficient_6=self.conic_coefficients2[6],
                conic_coefficient_7=self.conic_coefficients2[7],
                conic_coefficient_8=self.conic_coefficients2[8],
                conic_coefficient_9=self.conic_coefficients2[9],
                source_plane_distance = source_plane_distance_2,
                image_plane_distance = image_plane_distance_2,
                angles_respect_to = angles_respect_to_2,
                incidence_angle_deg = incidence_angle_deg_2,
                reflection_angle_deg = reflection_angle_deg_2,
                mirror_orientation_angle = mirror_orientation_angle_2,
                ))

        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                 str(exception),
                                 QMessageBox.Ok)


    def initializeTabs(self):
        self.tabs = oasysgui.tabWidget(self.mainArea)

        self.tab = [oasysgui.createTabPage(self.tabs, "Design parameters"),
                    oasysgui.createTabPage(self.tabs, "Output"),
                    oasysgui.createTabPage(self.tabs, "Mirror Profiles"),
        ]

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)


        self.design_output = oasysgui.textArea()
        tmp2 = oasysgui.widgetBox(self.tab[0], "Design output", addSpace=True, orientation="horizontal", height = self.IMAGE_HEIGHT-4) #, width=410)
        tmp2.layout().addWidget(self.design_output)

        self.shadow_output = oasysgui.textArea() #height=self.IMAGE_HEIGHT-5, width=400)
        tmp1 = oasysgui.widgetBox(self.tab[1], "System output", addSpace=True, orientation="horizontal", height = self.IMAGE_HEIGHT-4) #, width=410)
        tmp1.layout().addWidget(self.shadow_output)

        #
        self.plot_canvas = [None]

        self.plot_canvas[0] = oasysgui.plotWindow(roi=False, control=False, position=True)
        self.plot_canvas[0].setDefaultPlotLines(True)
        self.plot_canvas[0].setActiveCurveColor(color='blue')
        self.plot_canvas[0].setGraphYLabel("Z [nm]")
        self.plot_canvas[0].setGraphTitle("oe Profiles")
        self.plot_canvas[0].setInteractiveMode(mode='zoom')

        # self.plot_canvas[1] = oasysgui.plotWindow(roi=False, control=False, position=True)
        # self.plot_canvas[1].setDefaultPlotLines(True)
        # self.plot_canvas[1].setActiveCurveColor(color='blue')
        # self.plot_canvas[1].setGraphYLabel("Z [nm]")
        # self.plot_canvas[1].setGraphTitle("oe2 Profile")
        # self.plot_canvas[1].setInteractiveMode(mode='zoom')


        self.tab[2].layout().addWidget(self.plot_canvas[0])
        # self.tab[3].layout().addWidget(self.plot_canvas[1])

        self.tabs.setCurrentIndex(0)

    def check_fields(self):
        pass

    def writeStdOut(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

    def height(self, oe=1):

        if oe == 1:
            ccc = self.conic_coefficients1
        elif oe == 2:
            ccc = self.conic_coefficients2

        y = numpy.linspace(-self.y_length / 2, self.y_length / 2, self.npoints)
        x = 0

        aa = ccc[2]
        bb = ccc[4] * y + ccc[5] * x + ccc[8]
        cc = ccc[0] * x**2 + ccc[1] * y**2 + ccc[3] * x * y + \
            ccc[6] * x + ccc[7] * y + ccc[9]

        if aa != 0:
            discr = bb**2 - 4 * aa * cc + 0j
            s1 = (-bb + numpy.sqrt(discr)) / 2 / aa
            s2 = (-bb - numpy.sqrt(discr)) / 2 / aa
        else:
            s1 = -cc / bb
            s2 = s1

        return y, numpy.real(s1), numpy.real(s2)


    def plot_multi_data1D(self, x_list, y_list,
                    progressBarValue, tabs_canvas_index, plot_canvas_index,
                    title="", xtitle="",
                    ytitle="",
                    ytitles= [""],
                    colors = ['green'],
                    replace=True,
                    control=False,
                    xrange=None,
                    yrange=None,
                    symbol=['']):

        if len(y_list) != len(ytitles):
            ytitles = ytitles * len(y_list)

        if len(y_list) != len(colors):
            colors = colors * len(y_list)
        if len(y_list) != len(symbol):
            symbols = symbol * len(y_list)
        else:
            symbols = symbol

        if tabs_canvas_index is None: tabs_canvas_index = 0 #back compatibility?

        self.tab[tabs_canvas_index].layout().removeItem(self.tab[tabs_canvas_index].layout().itemAt(0))

        self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow(parent=None,
                                                                  backend=None,
                                                                  resetzoom=True,
                                                                  autoScale=False,
                                                                  logScale=True,
                                                                  grid=True,
                                                                  curveStyle=True,
                                                                  colormap=False,
                                                                  aspectRatio=False,
                                                                  yInverted=False,
                                                                  copy=True,
                                                                  save=True,
                                                                  print_=True,
                                                                  control=control,
                                                                  position=True,
                                                                  roi=False,
                                                                  mask=False,
                                                                  fit=False)


        self.plot_canvas[plot_canvas_index].setDefaultPlotLines(True)
        self.plot_canvas[plot_canvas_index].setActiveCurveColor(color=colors[0])
        self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)

        self.tab[tabs_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])


        for i in range(len(y_list)):
            # print(">>>>>>>>>>>>>>>>>>>> ADDING PLOT INDEX", i, x_list[i].shape, y_list[i].shape,ytitles[i],symbols[i],colors[i])
            self.plot_canvas[plot_canvas_index].addCurve(x_list[i], y_list[i],
                                         ytitles[i],
                                         xlabel=xtitle,
                                         ylabel=ytitle,
                                         symbol=symbols[i],
                                         color=colors[i])
        #
        self.plot_canvas[plot_canvas_index].getLegendsDockWidget().setFixedHeight(150)
        self.plot_canvas[plot_canvas_index].getLegendsDockWidget().setVisible(True)
        self.plot_canvas[plot_canvas_index].setActiveCurve(ytitles[0])
        self.plot_canvas[plot_canvas_index].replot()


        if xrange is not None:
            self.plot_canvas[plot_canvas_index].setGraphXLimits(xrange[0],xrange[1])
        if yrange is not None:
            self.plot_canvas[plot_canvas_index].setGraphYLimits(yrange[0],yrange[1])

        self.progressBarSet(progressBarValue)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = OWWolterCalculator()
    w.show()
    app.exec()
    w.saveSettings()



import os

try:
    from setuptools import find_packages, setup
except AttributeError:
    from setuptools import find_packages, setup

NAME = 'OASYS1-ESRF-Extensions'
VERSION = '0.0.73'
ISRELEASED = True

DESCRIPTION = 'OASYS extension for the ESRF'
README_FILE = os.path.join(os.path.dirname(__file__), 'README.md')
LONG_DESCRIPTION = open(README_FILE).read()
AUTHOR = 'Manuel Sanchez del Rio, Juan Reyes-Herrera, Luca Rebuffi'
AUTHOR_EMAIL = 'srio@esrf.eu'
URL = 'https://github.com/oasys-esrf-kit/OASYS1-ESRF-Extensions'
DOWNLOAD_URL = 'https://github.com/oasys-esrf-kit/OASYS1-ESRF-Extensions'
LICENSE = 'GPLv3'

KEYWORDS = (
    'raytracing',
    'simulator',
    'oasys1',
)

CLASSIFIERS = (
    'Development Status :: 4 - Beta',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Console',
    'Environment :: Plugins',
    'Programming Language :: Python :: 3',
    'Intended Audience :: Science/Research',
)

SETUP_REQUIRES = (
    'setuptools',
)

INSTALL_REQUIRES = (
    'setuptools',
    'oasys-barc4ro>=1.0.3',
    'pandas',
    'numba',
    'shadow4>=0.1.15',
    'xoppylib>=1.0.20',
    'crystalpy>=0.0.20',
)

PACKAGES = find_packages(exclude=('*.tests', '*.tests.*', 'tests.*', 'tests'))

PACKAGE_DATA = {
    "orangecontrib.esrf.oasys.widgets.extension":["icons/*.png", "icons/*.jpg"],
    "orangecontrib.esrf.wofry.widgets.extension":["icons/*.png", "icons/*.jpg"],
    "orangecontrib.esrf.xoppy.widgets.extension": ["icons/*.png", "icons/*.jpg"],
    "orangecontrib.esrf.syned.widgets.extension":["icons/*.png", "icons/*.jpg"],
    "orangecontrib.esrf.shadow.widgets.extension":["icons/*.png", "icons/*.jpg", "miscellanea/*.txt"],
    "orangecontrib.esrf.srw.widgets.extension":["icons/*.png", "icons/*.jpg"],
}

NAMESPACE_PACAKGES = ["orangecontrib",
                      "orangecontrib.esrf",
                      "orangecontrib.esrf.oasys",
                      "orangecontrib.esrf.wofry",
                      "orangecontrib.esrf.xoppy",
                      "orangecontrib.esrf.syned",
                      "orangecontrib.esrf.shadow",
                      "orangecontrib.esrf.srw",
                      "orangecontrib.esrf.oasys.widgets",
                      "orangecontrib.esrf.wofry.widgets",
                      "orangecontrib.esrf.xoppy.widgets",
                      "orangecontrib.esrf.syned.widgets",
                      "orangecontrib.esrf.shadow.widgets",
                      "orangecontrib.esrf.srw.widgets",
                      ]

ENTRY_POINTS = {
    'oasys.addons' : ("Oasys ESRF Extension = orangecontrib.esrf.oasys",
                      "Wofry ESRF Extension = orangecontrib.esrf.wofry",
                      "XOPPY ESRF Extension = orangecontrib.esrf.xoppy",
                      "Syned ESRF Extension = orangecontrib.esrf.syned",
                      "Shadow ESRF Extension = orangecontrib.esrf.shadow",
                      "SRW ESRF Extension = orangecontrib.esrf.srw",
                      ),
    'oasys.widgets' : (
        "Oasys ESRF Extension = orangecontrib.esrf.oasys.widgets.extension",
        "Wofry ESRF Extension = orangecontrib.esrf.wofry.widgets.extension",
        "XOPPY ESRF Extension = orangecontrib.esrf.xoppy.widgets.extension",
        "Syned ESRF Extension = orangecontrib.esrf.syned.widgets.extension",
        "Shadow ESRF Extension = orangecontrib.esrf.shadow.widgets.extension",
        "SRW ESRF Extension = orangecontrib.esrf.srw.widgets.extension",
    ),
    'oasys.menus' : ("esrfoasysmenu = orangecontrib.esrf.menu",)
}

if __name__ == '__main__':
    try:
        import PyMca5, PyQt4

        raise NotImplementedError("This version of ESRF Oasys Extensions doesn't work with Oasys1 beta.\nPlease install OASYS1 final release: http://www.elettra.eu/oasys.html")
    except:
        setup(
              name = NAME,
              version = VERSION,
              description = DESCRIPTION,
              long_description = LONG_DESCRIPTION,
              author = AUTHOR,
              author_email = AUTHOR_EMAIL,
              url = URL,
              download_url = DOWNLOAD_URL,
              license = LICENSE,
              keywords = KEYWORDS,
              classifiers = CLASSIFIERS,
              packages = PACKAGES,
              package_data = PACKAGE_DATA,
              #          py_modules = PY_MODULES,
              setup_requires = SETUP_REQUIRES,
              install_requires = INSTALL_REQUIRES,
              #extras_require = EXTRAS_REQUIRE,
              #dependency_links = DEPENDENCY_LINKS,
              entry_points = ENTRY_POINTS,
              namespace_packages=NAMESPACE_PACAKGES,
              include_package_data = True,
              zip_safe = False,
              )



import numpy
from orangecontrib.esrf.shadow.util.conic_penelope import ellipsoid, hyperboloid, paraboloid, rotate_and_shift_quartic, euler_rotation_matrix

from orangecontrib.esrf.shadow.util.conics_from_factory_parameters import ellipsoid as fac_ellipsoid
from orangecontrib.esrf.shadow.util.conics_from_factory_parameters import paraboloid as fac_paraboloid
from orangecontrib.esrf.shadow.util.conics_from_factory_parameters import hyperboloid as fac_hyperboloid
from orangecontrib.esrf.shadow.util.conics_from_factory_parameters import ken_ellipsoid, ken_hyperboloid, ken_paraboloid

def cyl(ccc):
    ccc1 = ccc.copy()
    ccc1[0] = 0
    ccc1[3] = 0
    ccc1[5] = 0
    ccc1[6] = 0
    return ccc1


def recipe1(
    p_ell=10.0,
    q_ell=3.0,
    distance=0.3,
    theta=0.003,
    theta_hyp=None,
    ratio_hyp=3.0, # ratio_hyp = q_hyp / p_ell > 1.0
    verbose=1,
                ):

    if theta_hyp is None: theta_hyp = theta

    q_hyp = q_ell - distance
    a_hyp = 0.5 * q_hyp * (1 - 1/ratio_hyp)
    p_hyp = q_hyp - 2 * a_hyp

    tkt_ell = ellipsoid(p_ell,q_ell,theta)
    tkt_hyp = hyperboloid(p_hyp,q_hyp,theta_hyp)

    if verbose:
        print("ell p,q", p_ell, q_ell)
        print("hyp p,q", p_hyp, q_hyp)

        print("position ell p,q", p_ell, distance / 2)
        print("position hyp p,q", distance / 2, p_hyp)

        print("length=", p_ell + distance + p_hyp)
        m_ell = (q_ell / p_ell)
        m_hyp = (p_hyp / q_hyp)
        print("magnification ell: ", m_ell)
        print("magnification hyp: ", m_hyp)
        print("magnification: ", m_ell * m_hyp)
        print("demagnification: ", 1 / (m_ell * m_hyp))

    return tkt_ell,tkt_hyp


def recipe2(
    p_ell=10.0,
    distance=0.3,
    p_hyp=0.9,
    theta=0.003,
    theta_hyp=None,
    # m_ell = 0.3,
    m_hyp=1/3,
    verbose=1,
            ):

    if theta_hyp is None: theta_hyp = theta

    q_hyp = p_hyp / m_hyp
    # a_hyp = 0.5 * (q_hyp - p_hyp)


    q_ell = q_hyp + distance

    m_ell = q_ell / p_ell


    if q_ell >= p_ell:
        raise Exception("must be: q_ell < p_ell")

    if q_ell >= p_ell:
        raise Exception("must be: q_ell < p_ell")

    if q_hyp >= q_ell:
        raise Exception("must be: q_hyp < q_ell")

    if q_hyp <= p_hyp:
        raise Exception("must be: q_hyp > p_hyp")

    tkt_ell = ellipsoid(p_ell,q_ell,theta)
    tkt_hyp = hyperboloid(p_hyp,q_hyp,theta_hyp)

    if verbose:
        print("distance:", distance)
        print("distance 2 (check):", q_ell - q_hyp)

        print("ell p,q", p_ell, q_ell)
        print("hyp p,q", p_hyp, q_hyp)

        print("position ell p,q", p_ell, distance / 2)
        print("position hyp p,q", distance / 2, p_hyp)

        print("length=", p_ell + distance + p_hyp)
        print("magnification ell: ", m_ell)
        print("magnification hyp: ", m_hyp)
        print("magnification: ", m_ell * m_hyp)
        print("demagnification: ", 1 / (m_ell * m_hyp))

    return tkt_ell,tkt_hyp

def recipe3(    #  common kick point
    p_ell = 100.0,
    q_ell = 10.0,
    p_hyp = 0.9,
    theta = 0.003,
    verbose = 1,
    method=0, # 0=Penelope, 1=Factory, 2=Ken
            ):

    q_hyp = q_ell

    if p_ell > 1e10: # use parabola
        if method == 0:
            tkt_ell = paraboloid(p_ell, q_ell, theta)
            ccc_ell = tkt_ell['ccc']
        elif method == 1:
            ccc_ell = fac_paraboloid(p_ell, q_ell, theta)
        elif method == 2:
            ccc_ell = ken_paraboloid(p_ell, q_ell, theta)

        f_ell = q_ell
    else:
        if method == 0:
            tkt_ell = ellipsoid(p_ell, q_ell, theta)
            ccc_ell = tkt_ell['ccc']
        elif method == 1:
            ccc_ell = fac_ellipsoid(p_ell, q_ell, theta)
        elif method == 2:
            ccc_ell = ken_ellipsoid(p_ell, q_ell, theta)

        f_ell = 1.0 / (1.0/p_ell + 1.0/q_ell)



    if method == 0:
        tkt_hyp = hyperboloid(p_hyp, q_hyp, theta)
        ccc_hyp = tkt_hyp['ccc']
    elif method == 1:
        ccc_hyp = fac_hyperboloid(p_hyp, q_hyp, theta)
    elif method == 2:
        ccc_hyp = ken_hyperboloid(p_hyp, q_hyp, theta)

    f_hyp = q_hyp / p_hyp

    # if p_hyp <= q_hyp:
    #     pass
    # else:
    #     raise Exception("must be: p_hyp <= q_hyp")


    if verbose:


        print("ell p,q", p_ell, q_ell)
        print("hyp p,q", p_hyp, q_hyp)

        print("length=", p_ell + p_hyp)
        print("f ell: ", f_ell)
        print("f hyp: ", f_hyp)
        print("f: ", f_ell * f_hyp)

    ccc_ell, ccc_hyp = numpy.array(ccc_ell), numpy.array(ccc_hyp)

    ccc_ell[numpy.argwhere(numpy.abs(ccc_ell) < 1e-14)] = 0
    ccc_hyp[numpy.argwhere(numpy.abs(ccc_hyp) < 1e-14)] = 0

    tkt_ell = {}
    tkt_ell['ccc'] = ccc_ell

    tkt_hyp = {}
    tkt_hyp['ccc'] = ccc_hyp

    return tkt_ell, tkt_hyp

def recipe4(    #  centered system parabola-hyperbola
    f11 = -0.00194644,
    f12 = 0.0,
    f21 = 1.905,
    f22 = 0.0,
    theta = 0.0159872,
    verbose = 1,
    method=0, # 0=Penelope, 1=Factory, 2=Ken
            ):

    if f12 != 0.0 or f22 != 0:
        raise Exception("Is your origin at the common focus?")

    p = numpy.abs(f11-f12)

    # intersection point at the parabola matching angle (https://doi.org/10.1107/S1600577522004593)
    c0 = (p / 2) / (numpy.tan(theta)) ** 2 - (p/2)
    # y^2 = 2px + p^2
    c1 = numpy.sqrt( 2 * p * c0 + p**2)

    ccc_centered_parabola = [1, 1, 0, 0, 0, 0, 0, 0, -2 * p, -p ** 2]  # normal incidence

    #
    # hyperbola
    #

    c = f21/2

    # get a from the hyperbola
    # (x-c)/a)^2 - (y/b)^2 = 1
    # b^2 = c^2 - a^2

    A = 1
    B = -c0**2 + 2 * c * c0 - 2 * c**2 - c1**2
    C = c**4 + c**2 * c0**2 - 2 * c**3 * c0
    S1 = numpy.sqrt( (-B + numpy.sqrt(B**2 - 4*A*C)) / (2 * A) )
    S2 = numpy.sqrt( (-B - numpy.sqrt(B ** 2 - 4 * A * C)) / (2 * A) )
    a = numpy.min((S1,S2))
    b = numpy.sqrt(c**2 - a**2)

    #
    # coeffs

    # Parabola
    # y^2 = p(2 * x + p) (Underwood)
    # x^2 + z^2 = 2 p y + p^2 (Shadow)
    ccc_centered_parabola = [1,1,0, 0,0,0,     0,0,-2*p, -p**2] # normal incidence
    ccc_centered_parabola = numpy.array(ccc_centered_parabola)

    # Hyperbola
    # (x-c)^2/a^2 - y^2/b^2 = 1 (Underwood)
    # (z-c)^2/a^2 - (y^2+x*2)/b^2 = 1 (Shadow)
    # # normal incidence (Underwood x->z, y->y  ->x)
    ccc_centered_hyperbola = [-1/b**2, -1/b**2, 1/a**2, \
                              0, 0, 0, \
                              0, 0, -2*c/a**2, \
                              (c/a)**2 - 1]
    ccc_centered_hyperbola = numpy.array(ccc_centered_hyperbola)


    if verbose:
        print("f11, f12", f11, f12)
        print("f21, f22", f21, f22)

        print("c0, c1: ", c0, c1)
        print("S1, S2: ", S1, S2)


        print("   theta grazing [deg]: ", theta * 180 / numpy.pi)
        print("   Parabola p [m]:", p)
        print("   Hyperbola a, b, c [m]: ", a, b, c)

        print("\n\nCalculated parameters: ")
        x_pmin = c0
        y_pmin = c1
        print("   ** Origin is at parabola focus (=far hyperbola focus)**")
        print("   Common point: x_pmin, y_pmin: ", x_pmin, y_pmin)
        print(" ")


        print("\n   normalized ccc_centered_parabola", ccc_centered_parabola / ccc_centered_parabola[0])
        print("   normalized ccc_centered_hyperbola", ccc_centered_hyperbola / ccc_centered_hyperbola[0])

        print("\n   ccc_centered_parabola", ccc_centered_parabola)
        print("   ccc_centered_hyperbola", ccc_centered_hyperbola)


    return  {'ccc':ccc_centered_parabola}, {'ccc':ccc_centered_hyperbola}

    # ccc_ell, ccc_hyp = numpy.array(ccc_ell), numpy.array(ccc_hyp)
    #
    # ccc_ell[numpy.argwhere(numpy.abs(ccc_ell) < 1e-14)] = 0
    # ccc_hyp[numpy.argwhere(numpy.abs(ccc_hyp) < 1e-14)] = 0
    #
    # tkt_ell = {}
    # tkt_ell['ccc'] = ccc_ell
    #
    # tkt_hyp = {}
    # tkt_hyp['ccc'] = ccc_hyp
    #
    # return tkt_ell, tkt_hyp
if __name__ == "__main__":

    if True:
        tkt_ell, tkt_hyp = recipe1(
            p_ell = 10.0,
            q_ell = 3.0,
            distance = 0.3,
            theta = 0.003,
            ratio_hyp = 3.0,
        )

        # print(tkt_ell)
        # print(tkt_hyp)

        print("\n\n>>>>>\n\n")
        # correct for incidence in the negative Y
        ccc1 = tkt_hyp['ccc']
        ccc2 = rotate_and_shift_quartic(ccc1, omega=0.0, theta=0.0, phi=numpy.pi, )
        print(ccc2)

        # """
        # [3703.7148148348147, 0.03333333333333333, 3703.7051501405344, 0.0, 11.965776068354531, 0.0, 0.0, 0.0, -102.56425641041795, 0.0]
        # [-45724.874256651594, -0.411522633747064, -45723.22816611661, 0.0, -548.6951988985766, 0.0, 0.0, 2.9816149549333204e-12, 740.7418518484149, -1.3322676295501878e-15]
        # """


    if False:
        # tkt_ell, tkt_hyp = recipe2(
        #     p_ell=10.0,
        #     distance=0.3,
        #     p_hyp=0.9,
        #     theta=0.003,
        #     # m_ell=0.3,
        #     m_hyp=1/3,
        #     verbose=1,
        # )

        tkt_ell, tkt_hyp = recipe2(
            p_ell=10.0,
            distance=0.5,
            p_hyp=2.5,
            theta=0.003,
            # m_ell=0.3,
            m_hyp=1/3,
            verbose=1,
        )

        print("\n\n>>>>>\n\n")
        # correct for incidence in the negative Y
        ccc1 = tkt_hyp['ccc']
        ccc2 = rotate_and_shift_quartic(ccc1, omega=0.0, theta=0.0, phi=numpy.pi, )
        print(ccc2)

        # """
        # [3703.7148148348147, 0.03333333333333333, 3703.7051501405344, 0.0, 11.965776068354531, 0.0, 0.0, 0.0, -102.56425641041795, 0.0]
        # [-45724.874256651594, -0.411522633747064, -45723.22816611661, 0.0, -548.6951988985766, 0.0, 0.0, 2.9816149549333204e-12, 740.7418518484149, -1.3322676295501878e-15]
        # """

    if False: # Underwood off-axis
        tkt_ell, tkt_hyp = recipe3( p_ell = 1e11,
                                    q_ell = 3.808047,
                                    p_hyp = 1.904995,
                                    theta = 0.0159872,
                                    verbose = 1,
                                    method = 0)
        ccc_ell = tkt_ell['ccc']
        ccc_hyp = tkt_hyp['ccc']

        print("\n\nNormalized Parabola: ",  ccc_ell/ccc_ell[0])
        print("Normalized Hyperbola: ", ccc_hyp/ccc_hyp[0])

        print("\n\nParabola: ",  ccc_ell)
        print("Hyperbola: ", ccc_hyp)


    if False: # Underwood centered
        tkt_ell, tkt_hyp = recipe4(verbose=1)
        ccc_ell = tkt_ell['ccc']
        ccc_hyp = tkt_hyp['ccc']

        # print("\n\nNormalized Parabola: ",  ccc_ell/ccc_ell[0])
        # print("Normalized Hyperbola: ", ccc_hyp/ccc_hyp[0])
        #
        # print("\n\nParabola: ",  ccc_ell)
        # print("Hyperbola: ", ccc_hyp)



from orangewidget.settings import Setting
from orangewidget import gui

from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import write_surface_file, read_surface_file
from oasys.util.oasys_objects import OasysSurfaceData

from syned.widget.widget_decorator import WidgetDecorator

from orangecontrib.wofry.util.wofry_objects import WofryData
from orangecontrib.wofry.widgets.gui.ow_optical_element import OWWOOpticalElement

from orangecontrib.esrf.wofry.util.thin_object import WOThinObject #TODO from wofryimpl....

class OWWOThinObject2D(OWWOOpticalElement):

    name = "ThinObject"
    description = "Wofry: Thin Object 2D"
    icon = "icons/thinb2d.png"
    priority = 206

    inputs = [("WofryData", WofryData, "set_input"),
              WidgetDecorator.syned_input_data()[0],
              ("Surface Data", OasysSurfaceData, "set_input_surface_data")
              ]


    material = Setting(1)
    refraction_index_delta = Setting(5.3e-7)
    att_coefficient = Setting(0.00357382)

    aperture_shape = Setting(0)
    aperture_dimension_v = Setting(100e-6)
    aperture_dimension_h = Setting(200e-6)


    file_with_thickness_mesh = Setting("<none>")

    def __init__(self):

        super().__init__(is_automatic=True, show_view_options=True, show_script_tab=True)

    def draw_specific_box(self):

        self.thinobject_box = oasysgui.widgetBox(self.tab_bas, "Thin Object Setting", addSpace=False, orientation="vertical",
                                           height=350)

        gui.comboBox(self.thinobject_box, self, "material", label="Lens material",
                     items=self.get_material_name(), callback=self.set_visible,
                     sendSelectedValue=False, orientation="horizontal")

        self.box_refraction_index_id = oasysgui.widgetBox(self.thinobject_box, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_refraction_index_id, self, "refraction_index_delta", "Refraction index delta",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("refraction_index_delta")

        self.box_att_coefficient_id = oasysgui.widgetBox(self.thinobject_box, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_att_coefficient_id, self, "att_coefficient", "Attenuation coefficient [m-1]",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("att_coefficient")



        filein_box = oasysgui.widgetBox(self.thinobject_box, "", addSpace=True,
                                        orientation="horizontal")  # width=550, height=50)
        self.le_beam_file_name = oasysgui.lineEdit(filein_box, self, "file_with_thickness_mesh",
                                                   "File with thickness mesh",
                                                   labelWidth=90, valueType=str, orientation="horizontal")
        gui.button(filein_box, self, "...", callback=self.selectFile)

        self.set_visible()


    def set_visible(self):
        self.box_refraction_index_id.setVisible(self.material in [0])
        self.box_att_coefficient_id.setVisible(self.material in [0])

    def selectFile(self):
        filename = oasysgui.selectFileFromDialog(self,
                previous_file_path=self.file_with_thickness_mesh, message="HDF5 Files (*.hdf5 *.h5 *.hdf)",
                start_directory=".", file_extension_filter="*.*")

        self.le_beam_file_name.setText(filename)


    def get_material_name(self, index=None):
        materials_list = ["External", "Be", "Al", "Diamond"]
        if index is None:
            return materials_list
        else:
            return materials_list[index]

    def get_optical_element(self):

        return WOThinObject(name=self.name,
                    file_with_thickness_mesh=self.file_with_thickness_mesh,
                    material=self.get_material_name(self.material),
                    refraction_index_delta=self.refraction_index_delta,
                    att_coefficient=self.att_coefficient,
                    )

    def check_data(self):
        super().check_data()
        congruence.checkFileName(self.file_with_thickness_mesh)

    def receive_specific_syned_data(self, optical_element):
        pass

    def set_input_surface_data(self, surface_data):
        if isinstance(surface_data, OasysSurfaceData):
           self.file_with_thickness_mesh = surface_data.surface_data_file
        else:
            raise Exception("Wrong surface_data")

    #
    # overwrite this method to add tab with thickness profile
    #

    def initializeTabs(self):
        size = len(self.tab)
        indexes = range(0, size)

        for index in indexes:
            self.tabs.removeTab(size-1-index)

        titles = ["Intensity","Phase","Thickness Profile"]
        self.tab = []
        self.plot_canvas = []

        for index in range(0, len(titles)):
            self.tab.append(gui.createTabPage(self.tabs, titles[index]))
            self.plot_canvas.append(None)

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)


    def do_plot_results(self, progressBarValue=80):
        super().do_plot_results(progressBarValue)
        if not self.view_type == 0:
            if not self.wavefront_to_plot is None:

                self.progressBarSet(progressBarValue)

                xx, yy, zz = self.get_optical_element().get_surface_thickness_mesh(self.input_data.get_wavefront())

                self.plot_data2D(data2D=1e6*zz.T,
                                 dataX=1e6*xx,
                                 dataY=1e6*xx,
                                 progressBarValue=progressBarValue,
                                 tabs_canvas_index=2,
                                 plot_canvas_index=2,
                                 title="O.E. Thickness profile in $\mu$m",
                                 xtitle="Horizontal [$\mu$m] ( %d pixels)"%(xx.size),
                                 ytitle="Vertical [$\mu$m] ( %d pixels)"%(yy.size))

                self.progressBarFinished()

            

if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication

    def get_example_wofry_data():
        from wofryimpl.propagator.light_source import WOLightSource
        from wofryimpl.beamline.beamline import WOBeamline
        from orangecontrib.wofry.util.wofry_objects import WofryData

        light_source = WOLightSource(dimension=2,
                                     initialize_from=0,
                                     range_from_h=-0.0003,
                                     range_to_h=0.0003,
                                     range_from_v=-0.0003,
                                     range_to_v=0.0003,
                                     number_of_points_h=400,
                                     number_of_points_v=200,
                                     energy=10000.0,
                                     )

        return WofryData(wavefront=light_source.get_wavefront(),
                           beamline=WOBeamline(light_source=light_source))



    a = QApplication(sys.argv)
    ow = OWWOThinObject2D()
    ow.file_with_thickness_mesh = "/home/srio/Downloads/SRW_M_thk_res_workflow_a_FC_CDn01.dat.h5"
    ow.set_input(get_example_wofry_data())


    ow.show()
    a.exec_()
    ow.saveSettings()




try:
    from shadow4.optical_surfaces.s4_conic import S4Conic # for comparison
    from shadow4tests.devel.wolter.conic_viewer import view_conic, compare_conics # for plot
except:
    pass

import numpy

from numpy import sin as Sin
from numpy import cos as Cos
from numpy import tan as Tan
from numpy import sqrt as Sqrt

def Cot(x):
    return 1/Tan(x)

def Sec(x):
    return 1/Cos(x)

def Csc(x):
    return 1/Sin(x)

# see conics_penelope_paraboloid_focusing.nb
def paraboloid(p=1e10, q=10,theta=3e-3):
    if p > q:
        return paraboloid_focusing(q=q, theta=theta)
    else:
        return paraboloid_collimating(p=p, theta=theta)

def paraboloid_focusing(q=10,theta=3e-3):
    return [1, Sin(theta)**2, Cos(theta)**2, 0, 2*Cos(theta)*Sin(theta), 0, 0, 0, -4*q*Sin(theta),0 ]

# see conics_penelope_paraboloid_collimating.nb
def paraboloid_collimating(p=10,theta=3e-3):
    return [1, Sin(theta) ** 2, Cos(theta) ** 2, 0, -2 * Cos(theta) * Sin(theta), 0, 0, 0, -4 * p * Sin(theta), 0]

# see conics_penelope_ellipsoid.nb
def ellipsoid(p=10,q=3,theta=3e-3):

    return [Csc(theta)**2/(p*q),
            1/(p*q),
            (-(p - q)**2 + (p + q)**2*Csc(theta)**2)/(p*q*(p + q)**2),
            0,
    (2*(p - q)*Sqrt(((p + q)**2*Cos(theta)**2)/(p**2 + q**2 + 2*p*q*Cos(2*theta)))*Sqrt(p**2 + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt(Csc(theta)**2/(p + q)**2))/(p*q*(p + q)),
            0,
            0,
    (4*(p - q)*(-(Sqrt(p*q)*Sqrt((p*q*Cos(theta)**2)/(p**2 + q**2 + 2*p*q*Cos(2*theta)))*Csc(theta)) +
        p*q*Sqrt(((p + q)**2*Cos(theta)**2)/(p**2 + q**2 + 2*p*q*Cos(2*theta)))*Sqrt(Csc(theta)**2/(p + q)**2)))/
        (p*q*(p + q)*Sqrt(p**2 + q**2 + 2*p*q*Cos(2*theta))*Sqrt(Csc(theta)**2/(p + q)**2)),
    -(Sqrt(Csc(theta)**2/(p + q)**2)*(-2*(p**2 - q**2)**2*Cot(theta)**2 +
    Csc(theta)**2*((p - q)**2*(p**2 + 6*p*q + q**2) + (p - q)**4*Cos(2*theta) +
       (8*(p*q)**1.5*Cos(theta)*Sqrt(((p + q)**2*Cos(theta)**2)/(p**2 + q**2 + 2*p*q*Cos(2*theta)))*Cot(theta))/
        (Sqrt((p*q*Cos(theta)**2)/(p**2 + q**2 + 2*p*q*Cos(2*theta)))*Sqrt(Csc(theta)**2/(p + q)**2))))*Sin(theta)**2)/
    (2.*p*q*(p**2 + q**2 + 2*p*q*Cos(2*theta))),
            0]

def hyperboloid(p=10,q=3,theta=3e-3):
    if p >= q:
        return hyperboloid_large_p(p=p,q=q,theta=theta)
    else:
        return hyperboloid_large_q(p=p, q=q, theta=theta)

def hyperboloid_large_p(p=10,q=3,theta=3e-3):
    if p < q:
        raise Exception("p<q")
    return [
        -(Csc(theta)**2/(p*q)),-(((p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*Csc(theta)**2)/
        (p*q*(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2))),
        (4*(p + q)**2 - ((p - q)**4*Csc(theta)**2*(1 + Csc(theta)**2))/(p*q))/((p - q)**2*(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)),0,
        (-2*(p + q)*(p**2 + q**2 - 2*p*q*Cos(2*theta))*Csc(theta)**2*Sqrt(1/(1 + (p + q)**2/((p - q)**2*(1 + Csc(theta)**2)))))/
        (p*(p - q)**3*q*Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))),0,0,
        -2*((-2*(p + q)*Sqrt(1/(1 + (p + q)**2/((p - q)**2*(1 + Csc(theta)**2)))))/
        ((p - q)*Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))) +
        (Sqrt(2)*(-p**2 + q**2)*Sqrt(-((p*q*(-3 + Cos(2*theta)))/(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))*
        Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))*Sin(theta))/
        (Sqrt(p*q)*(-2*(p**2 - p*q + q**2) + (p**2 + q**2)*Cos(2*theta)))),
        -((((p + q)**2*(p**2 + q**2 - 2*p*q*Cos(2*theta))*Csc(theta)*Sqrt(1/(2 + (2*(p + q)**2)/((p - q)**2*(1 + Csc(theta)**2))))*
        (4*Sqrt(p*q)*Sqrt(-((p*q*(-3 + Cos(2*theta)))/(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta)))) +
        (p - q)**2*Sqrt(((p - q)**2*(-3 + Cos(2*theta)))/(-2*(p**2 - p*q + q**2) + (p**2 + q**2)*Cos(2*theta)))*Csc(theta)*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))))/(p*q) +
        2*(4*(p + q)**2 - ((p - q)**4*Csc(theta)**2*(1 + Csc(theta)**2))/(p*q))*
        ((p + q)**2/(2.*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))) -
        Sqrt(2)*Sqrt(p*q)*Sqrt(-((p*q*(-3 + Cos(2*theta)))/(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))*
        Sqrt(1/(1 + (p + q)**2/((p - q)**2*(1 + Csc(theta)**2))))*Sin(theta)))/
        ((p - q)**2*(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2))),0]

def hyperboloid_large_q(p=3,q=10,theta=3e-3):
    if p > q:
        raise Exception("p>q")
    return [
        -(Csc(theta)**2/(p*q)),-(((p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*Csc(theta)**2)/
        (p*q*(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2))),
        (4*(p + q)**2 - ((p - q)**4*Csc(theta)**2*(1 + Csc(theta)**2))/(p*q))/((p - q)**2*(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)),0,
        (2*(p**2 + q**2 - 2*p*q*Cos(2*theta))*Sqrt(-((p*q*(-3 + Cos(2*theta)))/(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))*Csc(theta)**3*
        Sqrt((p + q)**2/(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)))/
        ((p - q)**2*(p*q)**1.5*Sqrt((4*(p**2 + q**2) + 2*(p - q)**2*Csc(theta)**2)/
        ((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))),0,0,
        -2*(-(Csc(theta)*(-4*p*q*Sqrt((p + q)**2/(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)) +
        ((p - q)*(p + q)*Csc(theta)**2)/
        (Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))))*
        (1 + Sin(theta)**2))/
        (2.*Sqrt(p*q)*(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)*
        Sqrt((p*q*(1 + Sin(theta)**2))/(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta)))) +
        ((p**2 + q**2 - 2*p*q*Cos(2*theta))*Sqrt(-((p*q*(-3 + Cos(2*theta)))/(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))*Csc(theta)**3*
        Sqrt((p + q)**2/(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2))*
        ((-2*p*q*(-3 + Cos(2*theta)))/
        ((p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))) +
        ((p - q)*(p + q)*Sqrt(((p + q)**2*Sin(theta)**2)/((p - q)**2 + 2*(p**2 + q**2)*Sin(theta)**2)))/
        Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))/
        (2.*(p - q)**2*(p*q)**1.5*Sqrt((4*(p**2 + q**2) + 2*(p - q)**2*Csc(theta)**2)/
        ((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta)))))),
        -2*((2*(p + q)*((p + q)**2/(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2))**1.5)/
        ((p - q)*Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))) +
        ((p - q)*(-3 + Cos(2*theta))*(-2*(p**2 - p*q + q**2) + (p**2 + q**2)*Cos(2*theta))*Csc(theta)**4*
        (p*(Sqrt((p + q)**2/(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)) -
        Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))))) +
        q*(Sqrt((p + q)**2/(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)) +
        Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*
        Sqrt((2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)/((p - q)**2*(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta)))))))/
        (Sqrt(p**2 - 4*p*q + q**2 + 2*p*q*Cos(2*theta))*(2*(p**2 + q**2) + (p - q)**2*Csc(theta)**2)**2)),0]


#
# Ken
#

def ken_paraboloid(p=1e11, q=10,theta=3e-3):
    if q < p:
        return ken_paraboloid_focusing(q=q, theta=theta)
    else:
        return ken_paraboloid_collimating(p=p, theta=theta)

def ken_paraboloid_focusing(q=10,theta=3e-3):
    return [1, Sin(theta)**2, Cos(theta)**2, 0, 2*Cos(theta)*Sin(theta), 0, 0, 0, -4*q*Sin(theta),0 ]

# see conics_penelope_paraboloid_collimating.nb
def ken_paraboloid_collimating(p=10,theta=3e-3):
    ccc = ken_paraboloid_focusing(q=p,theta=theta)
    ccc[4] = -ccc[4]
    return ccc

def ken_hyperboloid_large_q_old(p=3,q=10,theta=3e-3):
    c = Cos(theta)
    s = Sin(theta)
    return [
        0,
        -0.25 * s**2 * (q-p)**2,
        p*q - 0.25 * c**2 * (p+q)**2,
        0,
        0.5 * s * (q-p) * c * (p+q),
        0,
        0,
        0,
        s * (q-p) * p * q,
        0
    ]

def ken_ellipsoid(p=3,q=10,theta=3e-3):
    c = Cos(theta)
    s = Sin(theta)
    h = (p - q) * c
    return [
        (p+q)**2,
        (p+q)**2 * s**2,
        h**2 + 4 * p * q,
        0,
        2 * s * (p + q) * h,
        0,
        0,
        0,
        -4 * s * (p + q)  * p * q,
        0
    ]

def ken_hyperboloid(p=3,q=10,theta=3e-3):
    c = Cos(theta)
    s = Sin(theta)
    return [
        1,
        s**2,
        c**2 - 4 * p *q * s**2 / (q - p)**2,
        0,
        - 2 * s * c * (p + q) / (q - p),
        0,
        0,
        0,
        -4 * s * p * q  / (q - p),
        0
    ]

def ken_hyperboloid_large_q(p=3,q=10,theta=3e-3):
    return ken_hyperboloid(p,q,theta)

def ken_hyperboloid_large_p(p=3,q=10,theta=3e-3):
    return ken_hyperboloid(p,q,theta)

def ken_hyperboloid_large_p_old(p=3,q=10,theta=3e-3):
    return ken_hyperboloid_large_q_old(p,q,theta)

#
# tools
#
def cylinder(c_in):
    c_out = c_in.copy()
    c_out[0] = 0.0
    c_out[3] = 0.0
    c_out[5] = 0.0
    c_out[6] = 0.0
    return c_out

def normalize(c_in, index=0, clean=True):
    c_out = [0] * 10
    for i in range(10):
        c_out[i] = c_in[i] / c_in[index]
        if clean:
            if numpy.abs(c_out[i]) < 1e-15:
                c_out[i] = 0.0
    return c_out


#
# checks
#
#
#
#
def ellipsoid_check(ssour=10,simag=3,theta_grazing=3e-3, do_plot=False):


    ccc = S4Conic.initialize_as_ellipsoid_from_focal_distances(ssour, simag, theta_grazing,
                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
    print("ccc: ", ccc.get_coefficients())

    s5 = ellipsoid(ssour,simag,theta_grazing)

    c = ccc.get_coefficients()
    print("ccc: ", c)
    print("s5: ", s5)


    for i in range(10):
        print(i, c[i], s5[i])
        assert(numpy.abs(s5[i] - c[i]) < 1e-2)

    # view_conic(s5, x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1)
    if do_plot:
        compare_conics(s5, ccc.get_coefficients(), x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1,
                       titles=['s5','ccc'])

def parabola_check(ssour=10,simag=10,theta_grazing=3e-3, do_plot=False):


    ccc = S4Conic.initialize_as_paraboloid_from_focal_distances(ssour, simag, theta_grazing,
                                        cylindrical=0, cylangle=0.0, switch_convexity=0)
    print("ccc: ", ccc.get_coefficients())

    if simag < ssour:
        s5 = paraboloid_focusing(simag,theta_grazing)
    else:
        s5 = paraboloid_collimating(ssour, theta_grazing)


    print("ccc: ", ccc.get_coefficients())
    print("s5: ", s5)

    c = ccc.get_coefficients()
    for i in range(10):
        print(i, c[i], s5[i])

    for i in range(10):
        print(i, s5[i] , c[i])
        assert(numpy.abs(s5[i] - c[i]) < 1e-2)

    # view_conic(s5, x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1)
    if do_plot:
        compare_conics(s5, ccc.get_coefficients(), x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1,
                       titles=['s5','ccc'])

def hyperbola_check(ssour=10,simag=3,theta_grazing=3e-3, do_plot=False):

    if ssour > simag:
        s5 = hyperboloid_large_p(ssour, simag,theta_grazing)
        c = [-3703.714814834814, -0.033331264034642545, -3703.599850917718, 0, -41.26934605862947, 0, 0, 2.220446049250313e-16, -190.48238875690424, 0]

    else:
        s5 = hyperboloid_large_q(ssour, simag,theta_grazing)
        c = [-3703.714814834814, -0.033331264034642545, -3703.599850917718, 0, 41.26934605862948, 0, 0, 2.220446049250313e-16, 190.48238875690424, 0]


    print("ccc: ", c)
    print("s5: ", s5)

    for i in range(10):
        print(i, c[i], s5[i])

    for i in range(10):
        print(i, s5[i] , c[i])
        assert(numpy.abs(s5[i] - c[i]) < 1e-2)

    # view_conic(s5, x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1)
    if do_plot:
        compare_conics(s5, ccc.get_coefficients(), x_min=-0.01, x_max=0.01, y_min=-0.1, y_max=0.1,
                       titles=['s5','ccc'])

# adapted from s4_conic
def height(ccc,y=0,x=0,return_solution=0):
    aa = ccc[2]
    bb = ccc[4] * y + ccc[5] * x + ccc[8]
    cc = ccc[0] * x**2 + ccc[1] * y**2 + ccc[3] * x * y + \
        ccc[6] * x + ccc[7] * y + ccc[9]


    if aa != 0:
        discr = bb**2 - 4 * aa * cc + 0j
        print("delta: ", discr)
        s1 = (-bb + numpy.sqrt(discr)) / 2 / aa
        s2 = (-bb - numpy.sqrt(discr)) / 2 / aa

        # if return_solution == 0: # select the solution close to zero at pole
        #     if numpy.abs(s1).min() < numpy.abs(s2).min():
        #         ss = s1
        #     else:
        #         ss = s2
        # elif return_solution == 1:
        #     ss = s1
        # else:
        #     ss = s2
    else:
        # ss = -cc / bb
        s1 = -cc / bb
        s2 = -cc / bb

    return s1,s2

def plot_height(ccc, p=7, q=10, theta=3e-3, title=""):
    nx = 10
    ny = 20
    x0 = numpy.linspace(-0.1,0.1,nx)
    y0 = numpy.linspace(-0.5,0.5,ny)
    x = numpy.outer( x0, numpy.ones(ny) )
    y = numpy.outer( numpy.ones(nx), y0)
    z0,z1 = height(ccc, y=y, x=x)
    # print(x.shape, y.shape, z0.shape)

    z2 = height_ken(p, q, theta, y=y, x=x, sign1=+1.0)
    z3 = height_ken(p, q, theta, y=y, x=x, sign1=-1.0)

    if (z0.imag ** 2).sum() > 0: raise Exception("Delta < 0")
    if (z1.imag ** 2).sum() > 0: raise Exception("Delta < 0")

    from srxraylib.plot.gol import plot
    plot(
         y0, z0[nx // 2, :],
         y0, z1[nx // 2, :],
         y0, z2[nx // 2, :],
         y0, z3[nx // 2, :],
         legend=['solving eq 17 +', 'solving eq 17 -', 'Eq 11 +', 'Eq 11 -'],
         marker=[None,None,'+','+'],
         linestyle=[None,None,'',''],
         title=title,
         )

    plot(
         y0, z1[nx // 2, :],
         y0, z3[nx // 2, :],
         legend=['solving eq 17 -', 'Eq 11 -'],
         marker=[None,'+'],
         linestyle=[None,''],
         title=title,
         )

    plot(
         y0, z0[nx // 2, :],
         y0, z3[nx // 2, :],
         legend=['solving eq 17 +', 'Eq 11 -'],
         marker=[None,'+'],
         linestyle=[None,''],
         title=title,
         )

    return z0.real,z1.real

def height_ken(p, q, theta, y=0, x=0, sign1=+1.0):

    s = numpy.sin(theta)
    c = numpy.cos(theta)


    z = s*(2*p*q+c*(p+q)*y) + sign1 * numpy.sqrt(
        4 * s**2 * p * q * (p * q + c * (p + q) * y + y**2) - (c**2 * (p + q)**2 - 4 * p * q) * x**2
        )
    z /= (4 * s**2 * p * q / (p - q) - c**2 * (p - q))

    return z



if __name__ == "__main__":
    # print(ellipsoid(p=10,q=3,theta=3e-3))
    # print(paraboloid_focusing(q=10,theta=3e-3))
    # print(paraboloid_collimating(p=10, theta=3e-3))
    # print(hyperboloid_large_p(p=10, q=3, theta=3e-3))
    # print(hyperboloid_large_q(p=3, q=10, theta=3e-3))
    #
    # ellipsoid_check(ssour=10,simag=3,theta_grazing=3e-3, do_plot=False)
    # parabola_check(ssour=10e10,simag=10,theta_grazing=3e-3, do_plot=False)
    # parabola_check(ssour=10, simag=10e10, theta_grazing=3e-3, do_plot=False)
    # hyperbola_check(ssour=10, simag=3, theta_grazing=3e-3, do_plot=False)
    # hyperbola_check(ssour=3, simag=10, theta_grazing=3e-3, do_plot=False)

    # print("cylinder, p<q:")
    # print(normalize(cylinder(hyperboloid_large_q(p=3, q=10, theta=3e-3)), index=2))
    # print(normalize(ken_hyperboloid_large_q_old(p=3, q=10, theta=3e-3),index=2))
    #
    # print("cylinder, p>q:")
    # print(normalize(cylinder(hyperboloid_large_p(p=10, q=3, theta=3e-3)), index=2))
    # print(normalize(ken_hyperboloid_large_p_old(p=10, q=3, theta=3e-3),index=2))

    print("hyperboloid, p<q:")
    p, q, theta = 7, 10, 3e-1
    # p, q, theta = 0.900000, 2.700000, 0.003
    print(normalize(hyperboloid_large_q(p=p, q=q, theta=theta), index=0))
    print(ken_hyperboloid_large_q(p=p, q=q, theta=theta))
    ccc = ken_hyperboloid_large_q(p=p, q=q, theta=theta)
    s1, s2 = plot_height(ccc, p=p, q=q, theta=theta, title="p=%f, q=%f, theta=%f" % (p,q,theta))

    print("hyperboloid, p>q:")
    p, q, theta = 10, 7, 3e-1
    # p, q, theta = 2.7, 0.900000, 0.003
    print(normalize(hyperboloid_large_p(p=10, q=7, theta=3e-3), index=0))
    print(ken_hyperboloid_large_p(p=p, q=q, theta=theta))
    ccc = ken_hyperboloid_large_p(p=p, q=q, theta=theta)
    s1, s2 = plot_height(ccc, p=p, q=q, theta=theta, title="p=%f, q=%f, theta=%f" % (p,q,theta))



    # print("ellipsoid")
    # print(normalize(ellipsoid(p=10, q=3, theta=3e-3), index=0))
    # print(normalize(ken_ellipsoid(p=10, q=3, theta=3e-3), index=0))
    #
    # print("parabola")
    # print(normalize(paraboloid_focusing(q=10, theta=3e-3), index=0))
    # print(normalize(ken_paraboloid_focusing(q=10, theta=3e-3), index=0))